from __future__ import annotations

import json
import os
import time
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator
from warnings import warn

from myqueue.commands import Command, create_command
from myqueue.errors import parse_stderr
from myqueue.resources import Resources
from myqueue.states import State

if TYPE_CHECKING:
    from myqueue.schedulers import Scheduler

UNSPECIFIED = 'hydelifytskibadut'


class Task:
    """Task object.

    Parameters
    ----------

    cmd: :class:`myqueue.commands.Command`
        Command to be run.
    resources: Resources
        Combination of number of cores, nodename, number of processes
        and maximum time.
    deps: list of Path objects
        Dependencies.
    restart: int
        How many times to restart task.
    workflow: bool
        Task is part of a workflow.
    folder: Path
        Folder where task should run.
    creates: list of str
        Name of files created by task.
    """

    def __init__(self,
                 cmd: Command,
                 *,
                 resources: Resources,
                 deps: list[Path],
                 restart: int,
                 workflow: bool,
                 folder: Path,
                 creates: list[str],
                 notifications: str = '',
                 state: State = State.undefined,
                 id: int = 0,
                 error: str = '',
                 tqueued: float = 0.0,
                 trunning: float = 0.0,
                 tstop: float = 0.0,
                 user: str = ''):

        self.cmd = cmd
        self.resources = resources
        self.deps = deps
        self.restart = restart
        self.workflow = workflow
        self.folder = folder
        self.notifications = notifications
        self.creates = creates

        assert isinstance(state, State), state
        self.state = state
        self.id = id
        self.error = error

        # Timing:
        self.tqueued = tqueued
        self.trunning = trunning
        self.tstop = tstop

        self.user = user or os.environ.get('USER', 'root')

        self.dname = folder / cmd.name
        self.dtasks: list[Task] = []
        self._done: bool | None = None
        self.result = UNSPECIFIED

    @property
    def name(self) -> str:
        return f'{self.cmd.name}.{self.id}'

    def running_time(self, t: float = None) -> float:
        if self.state in ['CANCELED', 'queued', 'hold']:
            dt = 0.0
        elif self.state == 'running':
            t = t or time.time()
            dt = t - self.trunning
        else:
            dt = self.tstop - self.trunning
        return dt

    def words(self) -> list[str]:
        t = time.time()
        age = t - self.tqueued
        dt = self.running_time(t)

        info = []
        if self.restart:
            info.append(f'*{self.restart}')
        if self.deps:
            info.append(f'd{len(self.deps)}')
        if self.cmd.args:
            info.append(f'+{len(self.cmd.args)}')
        if self.notifications:
            info.append(self.notifications)

        return [str(self.id),
                str(self.folder) + '/',
                self.cmd.name.split('+', 1)[0],
                ' '.join(self.cmd.args),
                ','.join(info),
                str(self.resources),
                seconds_to_time_string(age),
                self.state.name,
                seconds_to_time_string(dt),
                self.error]

    def __str__(self) -> str:
        return ' '.join(self.words())

    def __repr__(self) -> str:
        return f'Task({self.cmd.name})'

    def order_key(self, column: str) -> Any:
        """ifnAraste"""
        if column == 'i':
            return self.id
        if column == 'f':
            return self.folder
        if column == 'n':
            return self.name
        if column == 'A':
            return len(self.cmd.args)
        if column == 'r':
            return self.resources.cores * self.resources.tmax
        if column == 'a':
            return self.tqueued
        if column == 's':
            return self.state.name
        if column == 't':
            return self.running_time()
        if column == 'e':
            return self.error
        raise ValueError(f'Unknown column: {column}!  '
                         'Must be one of i, f, n, a, I, r, A, s, t or e')

    def todict(self, root: Path = None) -> dict[str, Any]:
        folder = self.folder
        deps = self.deps
        if root:
            folder = folder.relative_to(root)
            deps = [dep.relative_to(root) for dep in self.deps]
        return {
            'id': self.id,
            'folder': str(folder),
            'cmd': self.cmd.todict(),
            'state': self.state.name,
            'resources': self.resources.todict(),
            'restart': self.restart,
            'workflow': self.workflow,
            'deps': [str(dep) for dep in deps],
            'notifications': self.notifications,
            'creates': self.creates,
            'tqueued': self.tqueued,
            'trunning': self.trunning,
            'tstop': self.tstop,
            'error': self.error,
            'user': self.user}

    def to_sql(self,
               root: Path) -> tuple[int, str, str, str, str,
                                    str, int, bool, str, float,
                                    str, str, float, float, float,
                                    str, str]:
        folder = str(self.folder.relative_to(root))
        if folder == '.':
            folder = './'
        else:
            folder = f'./{folder}/'
        return (self.id,
                folder,
                self.state.name[0],
                # str(self.dname.relative_to(root)),
                self.dname.name,
                json.dumps(self.cmd.todict()),
                json.dumps(self.resources.todict()),
                self.restart,
                self.workflow,
                ','.join(str(dep.relative_to(root)) for dep in self.deps),
                self.resources.weight,
                self.notifications,
                ','.join(self.creates),
                self.tqueued,
                self.trunning,
                self.tstop,
                self.error,
                self.user)

    @staticmethod
    def from_sql_row(row: tuple, root: Path) -> Task:
        (id, folder, state, name, cmd,
         resources, restart, workflow, deps, weight,
         notifications, creates, tqueued, trunning, tstop,
         error, user) = row
        resources = Resources(**json.loads(resources))
        assert resources.weight == weight
        return Task(id=id,
                    folder=root / folder,
                    state=State(state),
                    cmd=create_command(**json.loads(cmd)),
                    resources=resources,
                    restart=restart,
                    workflow=bool(workflow),
                    deps=[] if not deps else [root / dep
                                              for dep in deps.split(',')],
                    notifications=notifications,
                    creates=[] if not creates else creates.split(','),
                    tqueued=tqueued,
                    trunning=trunning,
                    tstop=tstop,
                    error=error,
                    user=user)

    @staticmethod
    def fromdict(dct: dict[str, Any], root: Path) -> Task:
        dct = dct.copy()

        # Backwards compatibility with version 2:
        if 'restart' not in dct:
            dct['restart'] = 0
        else:
            dct['restart'] = int(dct['restart'])

        dct.pop('diskspace', None)

        # Backwards compatibility:
        if 'creates' not in dct:
            dct['creates'] = []

        f = dct.pop('folder')
        if f.startswith('/'):
            # Backwards compatibility with version 5:
            folder = Path(f)
            deps = [Path(dep) for dep in dct.pop('deps')]
        else:
            folder = root / f
            deps = [root / dep for dep in dct.pop('deps')]

        id = int(dct.pop('id'))

        return Task(cmd=create_command(**dct.pop('cmd')),
                    resources=Resources(**dct.pop('resources')),
                    state=State[dct.pop('state')],
                    folder=folder,
                    deps=deps,
                    notifications=dct.pop('notifications', ''),
                    id=id,
                    **dct)

    def infolder(self, folder: Path, recursive: bool) -> bool:
        """Check if task runs inside a folder tree."""
        return folder == self.folder or (recursive and
                                         folder in self.folder.parents)

    def check_creates_files(self) -> bool:
        """Check if all files have been created."""
        if self.creates:
            for pattern in self.creates:
                if not any(self.folder.glob(pattern)):
                    return False
            return True
        return False

    def read_error_and_check_for_oom(self, scheduler: Scheduler) -> bool:
        """Check error message.

        Return True if out of memory.
        """
        self.error = '-'  # mark as already read

        path = scheduler.error_file(self)

        try:
            text = path.read_text()
        except (FileNotFoundError, UnicodeDecodeError):
            return False

        self.error, oom = parse_stderr(text)
        return oom

    def ideps(self, map: dict[Path, Task]) -> Iterator[Task]:
        """Yield task and its dependencies."""
        yield self
        for dname in self.deps:
            yield from map[dname].ideps(map)

    def submit(self, verbosity: int = 1, dry_run: bool = False) -> None:
        """Submit task.

        Parameters
        ----------

        verbosity: int
            Must be 0, 1 or 2.
        dry_run: bool
            Don't actually submit the task.
        """
        from myqueue.config import Configuration
        from myqueue.queue import Queue
        from myqueue.submitting import submit
        config = Configuration.read()
        with Queue(config, dry_run=dry_run) as queue:
            submit(queue, [self], verbosity=verbosity)

    def run(self) -> None:
        self.result = self.cmd.run()


def create_task(cmd: str,
                args: list[str] = [],
                *,
                resources: str = '',
                workflow: bool = False,
                name: str = '',
                deps: str | list[str] | Task | list[Task] = '',
                cores: int = 0,
                nodename: str = '',
                processes: int = 0,
                tmax: str = '',
                weight: float = -1.0,
                folder: str = '',
                restart: int = 0,
                creates: list[str] = []) -> Task:
    """Create a Task object.

    ::

        task = task('abc.py')

    Parameters
    ----------
    cmd: str
        Command to be run.
    args: list of str
        Command-line arguments or function arguments.
    resources: str
        Resources::

            'cores[:nodename][:processes]:tmax'

        Examples: '48:1d', '32:1h', '8:xeon8:1:30m'.  Can not be used
        togeter with any of "cores", "nodename", "processes" and "tmax".
    name: str
        Name to use for task.  Default is <cmd>[+<arg1>[_<arg2>[_<arg3>]...]].
    deps: str, list of str, Task object  or list of Task objects
        Dependencies.  Examples: "task1,task2", "['task1', 'task2']".
    cores: int
        Number of cores (default is 1).
    nodename: str
        Name of node.
    processes: int
        Number of processes to start (default is one for each core).
    tmax: str
        Maximum time for task.  Examples: "40s", "30m", "20h" and "2d".
    workflow: bool
        Task is part of a workflow.
    folder: str
        Folder where task should run (default is current folder).
    restart: int
        How many times to restart task.
    weight: float
        Weight of task.  See :ref:`task_weight`.
    creates: list of str
        Name of files created by task
        (can be both full filenames or patterns matching filenames).

    Returns
    -------
    Task
        Object representing the task.
    """

    path = Path(folder).absolute()

    dpaths = []
    if deps:
        if isinstance(deps, str):
            deps = deps.split(',')
        elif isinstance(deps, Task):
            deps = [deps]
        for dep in deps:
            if isinstance(dep, str):
                p = path / dep
                if '..' in p.parts:
                    p = p.parent.resolve() / p.name
                dpaths.append(p)
            else:
                dpaths.append(dep.dname)

    if '@' in cmd:
        # Old way of specifying resources:
        c, r = cmd.rsplit('@', 1)
        if r[0].isdigit():
            cmd = c
            resources = r
            warn(f'Please use resources={r!r} instead of deprecated '
                 f'...@{r} syntax!')

    command = create_command(cmd, args, name=name)

    res = Resources.from_args_and_command(
        cores, nodename, processes, tmax, weight, resources, command, path)

    return Task(command,
                resources=res,
                deps=dpaths,
                restart=restart,
                workflow=workflow,
                folder=path,
                creates=creates)


task = create_task


def seconds_to_time_string(n: float) -> str:
    """Convert number of seconds to string.

    >>> seconds_to_time_string(10)
    '0:10'
    >>> seconds_to_time_string(3601)
    '1:00:01'
    >>> seconds_to_time_string(24 * 3600)
    '1:00:00:00'
    """
    n = int(n)
    d, n = divmod(n, 24 * 3600)
    h, n = divmod(n, 3600)
    m, s = divmod(n, 60)
    if d:
        return f'{d}:{h:02}:{m:02}:{s:02}'
    if h:
        return f'{h}:{m:02}:{s:02}'
    return f'{m}:{s:02}'



import subprocess

import pytest
from myqueue.schedulers import SchedulerError
from myqueue.task import create_task


class Result:
    def __init__(self, stdout, stderr=b'', returncode=0):
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode


def run(commands,
        stdout=None,
        env=None,
        capture_output=None,
        input=None):
    """Fake subprocess.run() function."""
    # slurm:
    if any('FAIL' in arg for arg in commands):
        return Result(b'', b'FAIL', 1)
    if commands[0] == 'sbatch':
        return Result(b'ID: 42\n')
    if commands[0] == 'sacct':
        return Result(b'1K\n')
    if commands[0] == 'squeue':
        return Result(b'bla-bla\n1\n2\n')
    if commands[0] == 'scontrol':
        return
    if commands[0] == 'scancel':
        return

    # pbs:
    if commands[0] == 'qsub':
        return Result(b'42.hmmm\n')
    if commands[0] == 'qdel':
        return
    if commands[0] == 'qstat':
        return Result(b'bla-bla\n1.x abc\n2.x abc\n')

    # lsf:
    if commands[0] == 'bsub':
        return Result(b'bla-bla: j42.\n')
    if commands[0] == 'bkill':
        return
    if commands[0] == 'bjobs':
        return Result(b'bla-bla\n1 x\n2 abc\n')

    assert False, commands


@pytest.mark.parametrize('name', ['slurm', 'pbs', 'lsf'])
def test_scheduler_subprocess(monkeypatch, name):
    from myqueue.config import Configuration
    from myqueue.schedulers import get_scheduler

    monkeypatch.setattr(subprocess, 'run', run)

    config = Configuration(name)
    config.nodes = [('abc16', {'cores': 16, 'memory': '16G'}),
                    ('abc8', {'cores': 8, 'memory': '8G'})]
    scheduler = get_scheduler(config)
    t = create_task('x', resources='2:1h')
    scheduler.submit(t, dry_run=True, verbose=True)
    id = scheduler.submit(t)
    assert id == 42
    if name == 'slurm':
        scheduler.hold(42)
        scheduler.release_hold(42)
        assert scheduler.maxrss(1) == 1000
    scheduler.cancel(42)
    assert scheduler.get_ids() == {1, 2}

    t = create_task('FAIL', resources='2:1h')
    with pytest.raises(SchedulerError, match='FAIL'):
        scheduler.submit(t)



"""Simple caching function implementation using JSON."""
from __future__ import annotations
import json
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Sequence, TypeVar
from functools import lru_cache, wraps


T = TypeVar('T')


class CacheFileNotFoundError(FileNotFoundError):
    """JSON cache file not found."""


def json_cached_function(function: Callable[..., T],
                         name: str,
                         args: Sequence[Any],
                         kwargs: dict[str, Any]) -> Callable[..., T]:
    """Add file-caching to function.

    The decorated function will write its result in JSON format to a
    file called <name>.result.
    """
    path = Path(f'{name}.result')

    @wraps(function)
    def new_func(only_read_from_cache: bool = False) -> T:
        """A caching function.

        If *only_read_from_cache* is True then an CacheFileNotFoundError
        exception will be raised if the file does not exist.
        """
        if path.is_file():
            return decode(path.read_text(encoding='utf-8'))
        if only_read_from_cache:
            raise CacheFileNotFoundError
        result = function(*args, **kwargs)
        if mpi_world().rank == 0:
            path.write_text(encode(result), encoding='utf-8')
        return result

    return new_func


class MPIWorld:
    """A no-MPI implementation."""
    rank: int = 0


@lru_cache()
def mpi_world() -> MPIWorld:
    """Find and return a world object with a rank attribute."""
    import sys
    mod = sys.modules.get('mpi4py')
    if mod:
        return mod.MPI.COMM_WORLD  # type: ignore
    mod = sys.modules.get('_gpaw')
    if hasattr(mod, 'Communicator'):
        return mod.Communicator()  # type: ignore
    mod = sys.modules.get('_asap')
    if hasattr(mod, 'Communicator'):
        return mod.Communicator()  # type: ignore
    return MPIWorld()


class Encoder(json.JSONEncoder):
    """Encode complex, datetime, Path and ndarray objects.

    >>> import numpy as np
    >>> Encoder().encode(1+2j)
    '{"__complex__": [1.0, 2.0]}'
    >>> Encoder().encode(datetime(1969, 11, 11, 0, 0))
    '{"__datetime__": "1969-11-11T00:00:00"}'
    >>> Encoder().encode(Path('abc/123.xyz'))
    '{"__path__": "abc/123.xyz"}'
    >>> Encoder().encode(np.array([1., 2.]))
    '{"__ndarray__": [1.0, 2.0]}'
    """
    def default(self, obj: Any) -> Any:
        if isinstance(obj, complex):
            return {'__complex__': [obj.real, obj.imag]}

        if isinstance(obj, datetime):
            return {'__datetime__': obj.isoformat()}

        if isinstance(obj, Path):
            return {'__path__': str(obj)}

        if hasattr(obj, '__array__'):
            if obj.dtype == complex:
                dct = {'__ndarray__': obj.view(float).tolist(),
                       'dtype': 'complex'}
            else:
                dct = {'__ndarray__': obj.tolist()}
                if obj.dtype not in [int, float]:
                    dct['dtype'] = obj.dtype.name
            if obj.size == 0:
                dct['shape'] = obj.shape
            return dct

        return json.JSONEncoder.default(self, obj)


encode = Encoder().encode


def object_hook(dct: dict[str, Any]) -> Any:
    """Decode complex, datetime, Path and ndarray representations.

    >>> object_hook({'__complex__': [1.0, 2.0]})
    (1+2j)
    >>> object_hook({'__datetime__': '1969-11-11T00:00:00'})
    datetime.datetime(1969, 11, 11, 0, 0)
    >>> object_hook({'__path__': 'abc/123.xyz'})
    PosixPath('abc/123.xyz')
    >>> object_hook({'__ndarray__': [1.0, 2.0]})
    array([1., 2.])
    """
    data = dct.get('__complex__')
    if data is not None:
        return complex(*data)

    data = dct.get('__datetime__')
    if data is not None:
        return datetime.fromisoformat(data)

    data = dct.get('__path__')
    if data is not None:
        return Path(data)

    data = dct.get('__ndarray__')
    if data is not None:
        import numpy as np
        dtype = dct.get('dtype')
        if dtype == 'complex':
            array = np.array(data, dtype=float).view(complex)
        else:
            array = np.array(data, dtype=dtype)
        shape = dct.get('shape')
        if shape is not None:
            array.shape = shape
        return array

    return dct


def decode(text: str) -> Any:
    """Convert JSON to object(s)."""
    return json.loads(text, object_hook=object_hook)



"""Top level module definitions.

The version number and the submit() function is defined here.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from myqueue.task import Task


__version__ = '22.11.3'


def submit(*tasks: Task, verbosity: int = 1, dry_run: bool = False) -> None:
    """Submit tasks.

    Parameters
    ----------
    tasks: List of Task objects
        Tasks to submit.
    verbosity: int
        Must be 0, 1 or 2.
    dry_run: bool
        Don't actually submit the task.
    """
    from myqueue.queue import Queue
    from myqueue.config import Configuration
    from myqueue.submitting import submit as _submit

    config = Configuration.read()
    with Queue(config, dry_run=dry_run) as queue:
        _submit(queue, tasks, verbosity=verbosity)



import re
from pathlib import Path
from setuptools import setup, find_packages

txt = Path('myqueue/__init__.py').read_text()
version = re.search("__version__ = '(.*)'", txt).group(1)

long_description = Path('README.rst').read_text()

setup(name='myqueue',
      version=version,
      description='Simple job queue',
      long_description=long_description,
      author='J. J. Mortensen',
      author_email='jjmo@dtu.dk',
      url='https://myqueue.readthedocs.io/',
      packages=find_packages(),
      package_data={'myqueue': ['py.typed']},
      entry_points={'console_scripts': ['mq = myqueue.cli:main']},
      install_requires=['rich', 'networkx'],
      python_requires='>=3.7',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: '
          'GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: Unix',
          'Programming Language :: Python :: 3 :: Only',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Scientific/Engineering'])



from __future__ import annotations
from myqueue.workflow import run


def f(name='hi'):
    """Function that can fail with MEMORY error."""
    if name == 'oom':
        raise MemoryError


def workflow():
    """Long dependency chain."""
    d1 = run(function=f, name='1', args=['oom'])
    d2 = run(function=f, name='2', deps=[d1])
    d3 = run(function=f, name='3', deps=[d2])
    run(function=f, name='4', deps=[d3])



from __future__ import annotations
from myqueue.states import State
from myqueue.email import send_notification
from myqueue.pretty import pprint
from myqueue.utils import plural
from myqueue.queue import Queue
from myqueue.submitting import submit
from myqueue.hold import hold_or_release


def kick(queue: Queue, verbosity: int = 1) -> dict[str, int]:
    """Kick the system.

    * send email notifications
    * restart timed-out tasks
    * restart out-of-memory tasks
    * release/hold tasks to stay under *maximum_total_task_weight*
    """
    result = {}

    ndct = queue.config.notifications
    if ndct:
        notifications = send_notification(queue, **ndct)
        result['notifications'] = len(notifications)

    tasks = []
    sql = 'state IN ("T", "M") AND restart != 0 AND user = ?'
    for task in queue.tasks(sql, [queue.config.user]):
        nodes = queue.config.nodes or [('', {'cores': 1})]
        task.resources = task.resources.bigger(task.state, nodes)
        task.restart -= 1
        tasks.append(task)

    if tasks:
        if not queue.dry_run:
            with queue.connection as con:
                con.executemany('UPDATE tasks SET restart = 0 WHERE id = ?',
                                [(task.id,) for task in tasks])
        ids = list(queue.find_dependents(task.id for task in tasks))
        queue.cancel_dependents(ids)
        for id in ids:
            tasks.append(queue.tasks('id = ?', [id])[0])

        if queue.dry_run:
            pprint(tasks)
        else:
            if verbosity > 0:
                print('Restarting', plural(len(tasks), 'task'))
            for task in tasks:
                task.error = ''
                task.state = State.undefined
            submit(queue, tasks)
        result['restarts'] = len(tasks)

    result.update(hold_or_release(queue))

    return result



from __future__ import annotations
import os
import shlex
from pathlib import Path

import pytest  # type: ignore

from myqueue.cli import _main
from myqueue.config import Configuration
from myqueue.queue import Queue
from myqueue.task import Task
from myqueue.schedulers.test import TestScheduler


@pytest.fixture(scope='function')
def mq(tmpdir):
    dir = os.getcwd()
    os.chdir(tmpdir)
    yield MQ(Path(tmpdir))
    os.chdir(dir)


test_config = """\
config = {
    'scheduler': 'test',
    'notifications': {
        'email': 'me@myqueue.org',
        'host': 'smtp.myqueue.org',
        'username': 'me'},
    'mpiexec': 'echo',
    'maximum_total_task_weight': 2.5}
"""


class MQ:
    def __init__(self, dir):
        mqdir = dir / '.myqueue'
        mqdir.mkdir()
        (mqdir / 'config.py').write_text(test_config)
        self.config = Configuration.read()
        self.scheduler = TestScheduler(self.config)
        TestScheduler.current_scheduler = self.scheduler
        os.environ['MYQUEUE_TESTING'] = str(dir)

    def __call__(self, cmd: str, error: int = 0) -> None:
        args = shlex.split(cmd)
        if args and args[0][0] != '-' and args[0] != 'help':
            args[1:1] = ['--traceback']
        print(f'$ mq {cmd}')
        for i, arg in enumerate(args):
            if '*' in arg:
                args[i:i + 1] = sorted([str(p) for p in Path().glob(arg)])
                break
        err = _main(args)
        assert err == error

    def states(self) -> str:
        return ''.join(task.state.value
                       for task in mqlist(self.scheduler.config))

    def wait(self) -> str:
        while True:
            done = self.scheduler.kick()
            if done:
                break
        return self.states()


def mqlist(config) -> list[Task]:
    with Queue(config) as q:
        return q.select()



from __future__ import annotations
import subprocess

from myqueue.task import Task
from myqueue.schedulers import Scheduler, SchedulerError
from myqueue.utils import str2number


class LSF(Scheduler):
    def submit(self,
               task: Task,
               dry_run: bool = False,
               verbose: bool = False) -> int:
        nodelist = self.config.nodes
        nodes, nodename, nodedct = task.resources.select(nodelist)

        name = task.cmd.short_name
        hours, minutes = divmod(max(task.resources.tmax, 60) // 60, 60)

        bsub = ['bsub',
                '-J', name,
                '-W', f'{hours:02}:{minutes:02}',
                '-n', str(task.resources.cores),
                '-o', f'{task.folder}/{name}.%J.out',
                '-e', f'{task.folder}/{name}.%J.err',
                '-R', f'select[model == {nodename}]']

        mem = nodedct['memory']
        gbytes = int(str2number(mem) / 1_000_000_000 / nodedct['cores'])
        bsub += ['-R', f'rusage[mem={gbytes}G]']

        extra_args = self.config.extra_args + nodedct.get('extra_args', [])
        if extra_args:
            bsub += extra_args

        if task.dtasks:
            ids = ' && '.join(f'done({t.id})'
                              for t in task.dtasks)
            bsub += ['-w', f"{ids}"]

        # bsub += ['-R', f'span[hosts={nodes}]']
        bsub += ['-R', 'span[hosts=1]']

        cmd = str(task.cmd)
        if task.resources.processes > 1:
            cmd = 'mpiexec ' + cmd.replace('python3',
                                           self.config.parallel_python)

        home = self.config.home

        script = (
            '#!/bin/bash -l\n'
            'id=$LSB_JOBID\n'
            f'mq={home}/.myqueue/lsf-$id\n')

        script += self.get_venv_activation_line()

        script += (
            '(touch $mq-0 && \\\n'
            f' cd {str(task.folder)!r} && \\\n'
            f' {cmd} && \\\n'
            ' touch $mq-1) || \\\n'
            '(touch $mq-2; exit 1)\n')

        # print(' \\\n    '.join(bsub))
        # print(script)
        if dry_run:
            print(' \\\n    '.join(bsub))
            print(script)
            return 1

        p = subprocess.run(bsub,
                           input=script.encode(),
                           capture_output=True)
        if p.returncode:
            raise SchedulerError((p.stderr + p.stdout).decode())
        id = int(p.stdout.split()[1][1:-1].decode())
        return id

    def has_timed_out(self, task: Task) -> bool:
        path = self.error_file(task).expanduser()
        if path.is_file():
            task.tstop = path.stat().st_mtime
            lines = path.read_text().splitlines()
            for line in lines:
                if line.startswith('TERM_RUNLIMIT:'):
                    return True
        return False

    def cancel(self, id: int) -> None:
        subprocess.run(['bkill', str(id)])

    def get_ids(self) -> set[int]:
        p = subprocess.run(['bjobs'], stdout=subprocess.PIPE)
        queued = {int(line.split()[0].decode())
                  for line in p.stdout.splitlines()
                  if line[:1].isdigit()}
        return queued

    def get_config(self, queue: str = '') -> tuple[list[tuple[str, int, str]],
                                                   list[str]]:
        from collections import defaultdict

        cmd = ['nodestat', '-F', queue]
        p = subprocess.run(cmd, stdout=subprocess.PIPE)
        cores: dict[str, int] = {}
        memory: dict[str, list[str]] = defaultdict(list)
        for line in p.stdout.decode().splitlines():
            _, state, procs, _, name, mem, unit, *_ = line.split()
            if state == 'State':
                continue  # skip header
            if state == 'Down':
                continue
            cores[name] = int(procs.split(':')[1])
            memory[name].append(mem + unit)
        nodes = [
            (name, n, min(memory[name], key=str2number))
            for name, n in cores.items()]
        if queue:
            extra_args = ['-q', queue]
        else:
            extra_args = []
        return nodes, extra_args



from __future__ import annotations
from myqueue.commands import PythonModule


def test_funny_cli_args(capfd):
    arg = "--opt={...,x={'y':(1,2)}"
    cmd = PythonModule('myqueue.test.mod', [arg])
    cmd.run()
    args = capfd.readouterr().out.splitlines()
    assert args[1] == arg



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Generator, Callable
from math import inf
import rich.progress as progress
from rich.table import Table
from rich.console import Console

from myqueue import __version__
from myqueue.config import Configuration
from myqueue.queue import Queue
from myqueue.states import State


def info(queue: Queue,
         id: str = None,
         verbosity: int = 1) -> None:
    """Print information about MyQueue or a single task."""

    prnt = Console().print

    if id is None:
        table = Table(title=f'MyQueue-{__version__}')
        table.add_column(style='green')
        table.add_column('Path', style='cyan')
        table.add_row('Code', str(Path(__file__).parent))
        table.add_row('Root', str(queue.config.home / '.myqueue'))
        prnt(table)

        table = Table(title='Configuration')
        table.add_column('Key', style='green')
        table.add_column('Value', style='yellow')
        for key, value in queue.config.__dict__.items():
            if key == 'nodes':
                continue
            table.add_row(key, str(value))
        prnt(table)

        table = Table(title='Nodes')
        table.add_column('Name', style='green')
        table.add_column('Values', style='yellow')
        for name, dct in queue.config.nodes:
            table.add_row(name, str(dct))
        prnt(table)

        return

    task = queue.tasks('id = ?', [id])[0]
    prnt(json.dumps(task.todict(), indent='    '))
    if verbosity > 1:
        path = queue.scheduler.error_file(task)
        try:
            err = path.read_text()
        except FileNotFoundError:
            pass
        else:
            try:
                N = os.get_terminal_size().columns - 1
            except OSError:
                N = 70
            prnt(f'\nError file: {path}')
            prnt('v' * N)
            prnt(err)
            prnt('^' * N)


def info_all(start: Path) -> None:
    """Write information about all .myqueue folders."""
    dev = start.stat().st_dev
    nfolders = 0
    spinner = progress.Progress('[progress.description]{task.description}',
                                progress.SpinnerColumn('pong'),
                                transient=True)
    p = spinner.console.print

    with spinner:
        id = spinner.add_task('Searching', total=inf)
        for path in scan(start, dev, lambda: spinner.advance(id)):
            p(f'{path}:\n  ', end='')
            try:
                config = Configuration.read(path)
                nfolders += 1
            except FileNotFoundError as ex:
                p(f'[red]{ex}')
                continue
            with Queue(config, need_lock=False) as queue:
                from collections import defaultdict
                count: dict[str, int] = defaultdict(int)
                sql = 'SELECT state FROM tasks'
                for state, in queue.connection.execute(sql):
                    state = State(state).name
                    count[state] += 1
                count['total'] = sum(count.values())
                states = []
                for state, n in count.items():
                    if state == 'done':
                        state = '[green]done[/]'
                    elif state == 'running':
                        state = '[yellow]running[/]'
                    elif state.isupper():
                        state = f'[red]{state}[/]'
                    states.append(f'{state}: {n}')
                p(', '.join(states))
    print('Folders found:', nfolders)


def scan(path: Path,
         dev: int,
         spin: Callable[[], None]) -> Generator[Path, None, None]:
    """Scan for .myqueue folders.

    Only yield paths on same filesystem (dev).
    """

    with os.scandir(path) as entries:
        for entry in entries:
            spin()
            if entry.is_dir(follow_symlinks=False):
                if entry.name == '.myqueue':
                    yield path / entry.name
                elif (not entry.name.startswith(('.', '_')) and
                      entry.stat().st_dev == dev):
                    yield from scan(path / entry.name, dev, spin)



"""Queue class for interacting with the queue.

File format versions:

5)  Changed from mod:func to mod@func.
6)  Relative paths.
8)  Type of Task.id changed from int to str.
9)  Added "user".
10) Switched to sqlite3.
11) Renamed diskspace to weight.
"""
from __future__ import annotations

import sqlite3
import sys
import time
from pathlib import Path
from types import TracebackType
from typing import Iterator, Iterable, Sequence
from typing_extensions import LiteralString

from myqueue.config import Configuration
from myqueue.schedulers import Scheduler, get_scheduler
from myqueue.states import State
from myqueue.task import Task, create_task
from myqueue.utils import Lock, cached_property, normalize_folder
from myqueue.selection import Selection
from myqueue.migration import migrate

VERSION = 11

INIT = """\
CREATE TABLE tasks (
    id INTEGER PRIMARY KEY,
    folder TEXT,
    state CHARCTER,
    name TEXT,
    cmd TEXT,
    resources TEXT,
    restart INTEGER,
    workflow INTEGER,
    deps TEXT,
    weight REAL,
    notifications TEXT,
    creates TEXT,
    tqueued REAL,
    trunning REAL,
    tstop REAL,
    error TEXT,
    user TEXT);
CREATE TABLE dependencies (
    id INTEGER,
    did INTEGER,
    FOREIGN KEY (id) REFERENCES tasks(id),
    FOREIGN KEY (did) REFERENCES tasks(id));
CREATE TABLE meta (
    key TEXT,
    value TEXT);
CREATE INDEX folder_index on tasks(folder);
CREATE INDEX state_index on tasks(state);
CREATE INDEX dependincies_index1 on dependencies(id);
CREATE INDEX dependincies_index2 on dependencies(did)
"""


class DependencyError(Exception):
    """Bad dependency."""


class Queue:
    """Object for interacting with your .myqueue/queue.sqlite3 file"""
    def __init__(self,
                 config: Configuration = None,
                 *,
                 need_lock: bool = True,
                 dry_run: bool = False):
        self.need_lock = need_lock
        self.dry_run = dry_run
        self.config = config or Configuration('test')
        self.folder = self.config.home / '.myqueue'
        self.lock = Lock(self.folder / 'queue.sqlite3.myqueue.lock',
                         timeout=10.0)
        self._connection: sqlite3.Connection | None = None

    @cached_property
    def scheduler(self) -> Scheduler:
        """Scheduler object."""
        return get_scheduler(self.config)

    def __enter__(self) -> Queue:
        if self.need_lock:
            self.lock.acquire()
        else:
            try:
                self.lock.acquire()
            except PermissionError:
                pass  # it's OK to try to read without beeing able to write

        return self

    def __exit__(self,
                 type: Exception,
                 value: Exception,
                 tb: TracebackType) -> None:
        if self._connection:
            self._connection.close()
        self.lock.release()

    @property
    def connection(self) -> sqlite3.Connection:
        """Get or create a connection object."""
        if self._connection:
            return self._connection
        sqlfile = self.folder / 'queue.sqlite3'
        if self.lock.locked:
            self._connection = sqlite3.connect(sqlfile)
        else:
            self._connection = sqlite3.connect(f'file:{sqlfile}?mode=ro',
                                               uri=True)
        cur = self._connection.execute(
            'SELECT COUNT(*) FROM sqlite_master WHERE name="tasks"')

        if cur.fetchone()[0] == 0:
            self._initialize_db()
        else:
            version = int(
                self._connection.execute(
                    'SELECT value FROM meta where key="version"')
                .fetchone()[0])
            assert 11 <= version <= VERSION

        if self.lock.locked and not self.dry_run:
            self.process_change_files()
            self.check_for_timeout()
            self.check_for_oom()

        return self._connection

    def add(self, *tasks: Task) -> None:
        """Add tasks to database."""
        deps = []
        for task in tasks:
            for dep in task.dtasks:
                deps.append((task.id, dep.id))

        root = self.folder.parent
        q = ', '.join('?' * 17)
        with self.connection as con:
            con.executemany(
                f'INSERT INTO tasks VALUES ({q})',
                [task.to_sql(root) for task in tasks])
            con.executemany('INSERT INTO dependencies VALUES (?, ?)', deps)

    def sql(self,
            statement: LiteralString,
            args: list[str | int] = None) -> Iterator[tuple]:
        """Raw SQL execution."""
        return self.connection.execute(statement, args or [])

    def select(self, selection: Selection = None) -> list[Task]:
        """Create tasks from selection object."""
        root = self.folder.parent
        if selection:
            where, args = selection.sql_where_statement(root)
        else:
            where = ''
            args = []
        return self.tasks(where, args)

    def tasks(self,
              where: LiteralString,
              args: list[str | int] = None) -> list[Task]:
        """Create tasks from SQL WHERE statement."""
        root = self.folder.parent
        if where:
            sql = f'SELECT * FROM tasks WHERE {where}'
        else:
            sql = 'SELECT * FROM tasks'
        with self.connection:
            tasks = []
            for row in self.sql(sql, args or []):
                tasks.append(Task.from_sql_row(row, root))
        return tasks

    def _initialize_db(self) -> None:
        """Initialize tables and write version number."""
        assert self.lock.locked
        with self.connection:
            for statement in INIT.split(';'):
                self.connection.execute(statement)
            self.sql('INSERT INTO meta VALUES (?, ?)',
                     ['version', str(VERSION)])

        jsonfile = self.folder / 'queue.json'
        if jsonfile.is_file():
            migrate(jsonfile, self.connection)

    def find_dependents(self,
                        ids: Iterable[int],
                        known: dict[int, list[int]] = None) -> Iterator[int]:
        """Yield dependents."""
        if known is None:
            known = {}
        result = set()
        for id in ids:
            if id in known:
                result.update(known[id])
            else:
                dependents = [
                    id for id, in self.sql(
                        'SELECT id FROM dependencies WHERE did = ?', [id])]
                known[id] = dependents
                result.update(dependents)
        if result:
            yield from result
            yield from self.find_dependents(result, known)

    def cancel_dependents(self, ids: Iterable[int]) -> None:
        """Set state of dependents to CANCELED."""
        if self.dry_run:
            return
        t = time.time()
        args = [(t, id) for id in self.find_dependents(ids)]
        with self.connection as con:
            con.executemany(
                'UPDATE tasks SET state = "C", tstop = ? WHERE id = ?', args)

    def remove(self, ids: Iterable[int]) -> None:
        """Remove tasks."""
        if self.dry_run:
            return
        ids = list(ids)
        self.cancel_dependents(ids)
        args = [[id] for id in ids]
        with self.connection as con:
            con.executemany('DELETE FROM dependencies WHERE id = ?', args)
            con.executemany('DELETE FROM dependencies WHERE did = ?', args)
            con.executemany('DELETE FROM tasks WHERE id = ?', args)

    def check_for_timeout(self) -> None:
        """Find "running" tasks that are actually timed out."""
        t = time.time()

        timeouts = []
        for task in self.tasks('state = "r"'):
            delta = t - task.trunning - task.resources.tmax
            if delta > 0:
                if self.scheduler.has_timed_out(task) or delta > 1800:
                    timeouts.append(task.id)

        with self.connection:
            self.connection.executemany(
                'UPDATE tasks SET state = "T", tstop = ? WHERE id = ?',
                [(t, id) for id in timeouts])
        self.cancel_dependents(timeouts)

    def check_for_oom(self) -> None:
        """Fin out-of-memory tasks."""
        args = []
        for task in self.tasks('state = "F" AND error = ""'):
            oom = task.read_error_and_check_for_oom(self.scheduler)
            args.append(('M' if oom else 'F', task.error, task.id))
        with self.connection:
            self.connection.executemany(
                'UPDATE tasks SET state = ?, error = ? WHERE id = ?', args)

    def process_change_files(self) -> None:
        """Process state-change files from running tasks."""
        paths = list(self.folder.glob('*-*-*'))
        states = {0: State.running,
                  1: State.done,
                  2: State.FAILED,
                  3: State.TIMEOUT}
        files = []
        for path in paths:
            id, state = (int(x) for x in path.name.split('-')[1:])
            files.append((path.stat().st_ctime, id, state, path))

        for ctime, id, state, path in sorted(files):
            self.update_one_task(id, states[state], ctime, path)

    def update_one_task(self,
                        id: int,
                        newstate: State,
                        ctime: float,
                        path: Path) -> None:
        """Update single task."""
        try:
            (user,), = self.sql('SELECT user FROM tasks WHERE id = ?', [id])
        except ValueError:
            print(f'No such task: {id}, {newstate}', file=sys.stderr)
            path.unlink()
            return

        if user != self.config.user:
            return

        if newstate == 'done':
            with self.connection as con:
                con.execute('DELETE FROM dependencies WHERE did = ?', [id])
            with self.connection as con:
                con.execute(
                    'UPDATE tasks SET state = "d", tstop = ? WHERE id = ?',
                    [ctime, id])

        elif newstate == 'running':
            with self.connection as con:
                con.execute(
                    'UPDATE tasks SET state = "r", trunning = ? WHERE id = ?',
                    [ctime, id])

        else:
            assert newstate in ['FAILED', 'TIMEOUT', 'MEMORY']
            self.cancel_dependents([id])
            with self.connection as con:
                con.execute(
                    'UPDATE tasks SET state = ?, tstop = ? WHERE id = ?',
                    [newstate.value, ctime, id])

        path.unlink()


def sort_out_dependencies(tasks: Sequence[Task],
                          queue: Queue = None,
                          done: list[Task] = None) -> None:
    """Get dependencies ready for submitting."""
    root = queue.config.home if queue is not None else Path('.').absolute()

    name_to_task = {str(task.dname.relative_to(root)): task
                    for task in tasks}

    name_to_id_and_state: dict[str, tuple[int, str]] = {}

    if done is not None:
        for task in done:
            name_to_id_and_state[str(task.dname.relative_to(root))] = (0, 'd')

    for task in tasks:
        task.dtasks = []
        deps = []
        for dname in task.deps:
            name = str(dname.relative_to(root))
            dtask = name_to_task.get(name)
            if dtask is None:
                id, state = name_to_id_and_state.get(name, (-1, ''))
                if id == -1:
                    assert queue is not None
                    rows = queue.sql(
                        'SELECT id, state FROM tasks '
                        'WHERE name = ? AND folder = ?',
                        [dname.name, normalize_folder(task.folder, root)])
                    id, state = max(rows, default=(-1, ''))
                    if id == -1:
                        raise DependencyError(f"Can't find {name}")
                    name_to_id_and_state[name] = id, state
                if state in 'qhr':
                    dtask = create_task('dummy')
                    dtask.id = id
                    dtask.state = State(state)
                elif state == 'd':
                    continue
                else:
                    raise DependencyError(f'Bad state ({state}): {name}')

            task.dtasks.append(dtask)
            deps.append(dname)
        task.deps = deps


def dump_db(path: Path) -> None:
    """Pretty-print content of sqlite3 db file."""
    from rich.table import Table
    from rich.console import Console
    prnt = Console().print
    db = sqlite3.connect(path)
    table = Table(title=str(path))
    columns = [line.strip().split()[0]
               for line in INIT.split(';', maxsplit=1)[0].splitlines()[1:]]
    for name in columns:
        table.add_column(name)
    for row in db.execute('SELECT * from tasks'):
        table.add_row(*[str(x) for x in row])
    prnt(table)

    table = Table(title='dependencies')
    table.add_column('id')
    table.add_column('did')
    for row in db.execute('SELECT * from dependencies'):
        table.add_row(*[str(x) for x in row])
    prnt(table)

    table = Table(title='meta')
    table.add_column('key')
    table.add_column('value')
    for row in db.execute('SELECT * from meta'):
        table.add_row(*[str(x) for x in row])
    prnt(table)


if __name__ == '__main__':
    dump_db(Path(sys.argv[1]))



from myqueue.config import Configuration
from myqueue.schedulers.scheduler import Scheduler


class SchedulerError(Exception):
    """Error from scheduler command."""


def get_scheduler(config: Configuration) -> Scheduler:
    """Create scheduler from config object."""
    name = config.scheduler.lower()
    if name == 'test':
        from myqueue.schedulers.test import TestScheduler
        assert TestScheduler.current_scheduler is not None
        return TestScheduler.current_scheduler
    if name == 'local':
        from myqueue.schedulers.local import LocalScheduler
        return LocalScheduler(config)
    if name == 'slurm':
        from myqueue.schedulers.slurm import SLURM
        return SLURM(config)
    if name == 'pbs':
        from myqueue.schedulers.pbs import PBS
        return PBS(config)
    if name == 'lsf':
        from myqueue.schedulers.lsf import LSF
        return LSF(config)
    raise ValueError(f'Unknown scheduler: {name}')



from __future__ import annotations
from enum import Enum


class State(Enum):
    """Task-state enum.

    The following 9 states are defined:

    >>> for state in State:
    ...     state
    <State.undefined: 'u'>
    <State.queued: 'q'>
    <State.hold: 'h'>
    <State.running: 'r'>
    <State.done: 'd'>
    <State.FAILED: 'F'>
    <State.TIMEOUT: 'T'>
    <State.MEMORY: 'M'>
    <State.CANCELED: 'C'>

    >>> State.queued == State.queued
    True
    >>> State.queued == 'queued'
    True
    >>> State.queued == 'queue'
    Traceback (most recent call last):
      ...
    TypeError: Unknown state: queue
    >>> State.queued == 117
    False
    >>> State.done in {'queued', 'running'}
    False
    """

    undefined = 'u'
    queued = 'q'
    hold = 'h'
    running = 'r'
    done = 'd'
    FAILED = 'F'
    TIMEOUT = 'T'
    MEMORY = 'M'
    CANCELED = 'C'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, State):
            return self.name == other.name
        if isinstance(other, str):
            if other in State.__members__:
                return self.name == other
            raise TypeError(f'Unknown state: {other}')
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.name)

    def __str__(self) -> str:
        return self.name

    def is_bad(self) -> bool:
        """Return true for FAILED, TIMEOUT, MEMORY and CANCELED.

        >>> State.running.is_bad()
        False
        """
        return self.name.isupper()

    def is_active(self) -> bool:
        """Return true for queued, hold, running.

        >>> State.hold.is_active()
        True
        """
        return self.value in 'qhr'

    @staticmethod
    def str2states(s: str) -> set[State]:
        """Convert single character state string to set of State objects.

        >>> names = [state.name for state in State.str2states('rdA')]
        >>> sorted(names)
        ['CANCELED', 'FAILED', 'MEMORY', 'TIMEOUT', 'done', 'running']
        >>> State.str2states('x')
        Traceback (most recent call last):
          ...
        ValueError: Unknown state: x.  Must be one of q, ..., a or A.
        """
        states: set[State] = set()
        for c in s:
            if c == 'a':
                states.update([State.queued,
                               State.hold,
                               State.running,
                               State.done])
            elif c == 'A':
                states.update([State.FAILED,
                               State.CANCELED,
                               State.TIMEOUT,
                               State.MEMORY])
            else:
                try:
                    states.add(State(c))
                except ValueError:
                    raise ValueError(
                        'Unknown state: ' + s +
                        '.  Must be one of q, h, r, d, F, C, T, M, a or A.')
        return states



"""Resource class to handle resource requirements: time, cores, processes."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Tuple, Dict

if TYPE_CHECKING:
    from myqueue.commands import Command

from myqueue.states import State

Node = Tuple[str, Dict[str, Any]]

RESOURCES_URL = (
    'https://myqueue.readthedocs.io/en/latest/documentation.html#resources')


def seconds_to_short_time_string(n: float) -> str:
    """Convert seconds to time string.

    >>> seconds_to_short_time_string(42)
    '42s'
    >>> seconds_to_short_time_string(7200)
    '2h'
    """
    n = int(n)
    for s, t in [('d', 24 * 3600),
                 ('h', 3600),
                 ('m', 60),
                 ('s', 1)]:
        if n % t == 0:
            break

    return f'{n // t}{s}'


def T(t: str) -> int:
    """Convert string to seconds."""
    return {'s': 1,
            'm': 60,
            'h': 3600,
            'd': 24 * 3600}[t[-1]] * int(t[:-1])


class Resources:
    """Resource description."""
    def __init__(self,
                 cores: int = 0,
                 nodename: str = '',
                 processes: int = 0,
                 tmax: int = 0,
                 weight: float = -1.0):
        """Resource object.

        cores: int
            Number of cores.
        nodename: str
            Name of node.
        processes: int
            Number of processes to start.
        tmax: str
            Maximum time for task.  Examples: "40s", "30m", "20h" and "2d".
        weight: float
            Weight of task.  See :ref:`task_weight`.
        """
        self.cores = cores or 1
        self.nodename = nodename
        self.tmax = tmax or 600  # seconds
        self.weight = weight

        if processes == 0:
            self.processes = self.cores
        else:
            self.processes = processes

    def set_default_weight(self, weight: float) -> None:
        if self.weight == -1.0:
            self.weight = weight

    @staticmethod
    def from_string(s: str) -> Resources:
        """Create Resource object from string.

        >>> Resources.from_string('16:1:xeon8:2h')
        Resources(cores=16, processes=1, tmax=7200, nodename='xeon8')
        >>> Resources.from_string('16:1m')
        Resources(cores=16, tmax=60)
        >>> r = Resources.from_string('16:1m:25')
        >>> r
        Resources(cores=16, tmax=60, weight=25.0)
        >>> print(r)
        16:1m:25
        """
        nodename = ''
        processes = 0
        weight = -1.0
        try:
            p1, *parts = s.split(':')
            cores = int(p1)
            if parts[-1][-1] not in 'smhd':
                weight = float(parts.pop())
            tmax = T(parts.pop())
            for p in parts:
                if p.isdigit():
                    processes = int(p)
                else:
                    nodename = p
        except ValueError as ex:
            raise ValueError(
                f'Bad resource string: {s!r}.  See {RESOURCES_URL}') from ex

        return Resources(cores, nodename, processes, tmax, weight)

    @staticmethod
    def from_args_and_command(cores: int = 0,
                              nodename: str = '',
                              processes: int = 0,
                              tmax: str = '',
                              weight: float = -1.0,
                              resources: str = '',
                              command: Command = None,
                              path: Path = None) -> Resources:
        all_defaults = (cores == 0 and
                        nodename == ''
                        and processes == 0
                        and tmax == '' and
                        weight == -1.0)
        if all_defaults:
            if resources:
                return Resources.from_string(resources)
            assert command is not None and path is not None
            res = command.read_resources(path)
            if res is not None:
                return res
        else:
            assert resources == ''

        return Resources(cores, nodename, processes, T(tmax or '10m'), weight)

    def __str__(self) -> str:
        s = str(self.cores)
        if self.nodename:
            s += ':' + self.nodename
        if self.processes != self.cores:
            s += ':' + str(self.processes)
        s += ':' + seconds_to_short_time_string(self.tmax)
        if self.weight > 0.0:
            s += f':{int(self.weight)}'
        return s

    def __repr__(self) -> str:
        args = ', '.join(f'{key}={value!r}'
                         for key, value in self.todict().items())
        return f'Resources({args})'

    def todict(self) -> dict[str, Any]:
        """Convert to dict."""
        dct: dict[str, float | int | str] = {'cores': self.cores}
        if self.processes != self.cores:
            dct['processes'] = self.processes
        if self.tmax != 600:
            dct['tmax'] = self.tmax
        if self.nodename:
            dct['nodename'] = self.nodename
        if self.weight != -1.0:
            dct['weight'] = self.weight
        return dct

    def bigger(self,
               state: State,
               nodelist: list[Node],
               maxtmax: int = 2 * 24 * 3600) -> Resources:
        """Create new Resource object with larger tmax or more cores.

        >>> nodes = [('node1', {'cores': 8})]
        >>> r = Resources(tmax=100, cores=8)
        >>> r.bigger(State.TIMEOUT, nodes)
        Resources(cores=8, tmax=200)
        >>> r.bigger(State.MEMORY, nodes)
        Resources(cores=16, tmax=100)
        """
        new = Resources(**self.todict())
        if state == 'TIMEOUT':
            new.tmax = int(min(self.tmax * 2, maxtmax))
        elif state == 'MEMORY':
            coreslist = sorted({dct['cores'] for name, dct in nodelist})
            nnodes = 1
            while True:
                for c in coreslist:
                    cores = nnodes * c
                    if cores > self.cores:
                        break
                else:
                    nnodes += 1
                    continue
                break
            if self.processes == self.cores:
                new.processes = cores
            new.cores = cores
        else:
            raise ValueError
        return new

    def select(self, nodelist: list[Node]) -> tuple[int, str, dict[str, Any]]:
        """Select appropriate node.

        >>> nodes = [('node1', {'cores': 16}),
        ...          ('node2', {'cores': 8}),
        ...          ('fatnode2', {'cores': 8})]
        >>> Resources(cores=24).select(nodes)
        (3, 'node2', {'cores': 8})
        >>> Resources(cores=32).select(nodes)
        (2, 'node1', {'cores': 16})
        >>> Resources(cores=32, nodename='fatnode2').select(nodes)
        (4, 'fatnode2', {'cores': 8})
        >>> Resources(cores=1).select(nodes)
        (1, 'node2', {'cores': 8})
        >>> Resources(cores=32, nodename='node3').select(nodes)
        Traceback (most recent call last):
            ...
        ValueError: No such node: node3
        """
        if self.nodename:
            for name, dct in nodelist:
                if name == self.nodename:
                    break
            else:  # no break
                raise ValueError(f'No such node: {self.nodename}')
        else:
            for name, dct in nodelist:
                if self.cores % dct['cores'] == 0:
                    break
            else:  # no break
                node = min(nodelist, key=lambda node: node[1]['cores'])
                name, dct = node

        nodes, rest = divmod(self.cores, dct['cores'])
        if rest:
            nodes += 1

        return nodes, name, dct



import os

from myqueue.queue import Queue


def test_no_such_task(tmp_path):
    os.chdir(tmp_path)
    mq = tmp_path / '.myqueue'
    mq.mkdir()
    p = mq / 'test-42-2'
    p.write_text('')
    with Queue() as q:
        q.connection  # this will trigger q.update_one_task()
    assert not p.is_file()
